/******************************************************************************
* Dis8051 v2.0                                                                *
*  v1.1 fixed the problem with backward short jumps.                          *
*  v1.2 can load binary image or Intel .HEX files as well.                    *
*  v1.3 fixed the bug where LCALL was not recognised.                         *
*  v1.4 prints out the address before each line, names of interrupt vectors   *
*    and the version in the .ASM file, and the directory on the help screen.  *
*  v1.5 deals with proper labelling of LCALL and LJMP, and flags down ACALLs  *
*    and AJMPs as warnings. Also, a bug was fixed where it prints warnings,   *
*    and the dodgy <cr>'s in the dest file were also fixed.                   *
*  v1.6 deals properly with AJMP and ACALL, and introduces a top address      *
*    to allow for LUT's without causing problems.                             *
*                                                                             *
*  v1.9 - Development version for 2.0                                         *
*    a) Allows selective printing of NOPs and Addresses.                      *
*    b) Fixed the bit in PrintArg(), for printing unnamed direct addresses.   *
*    c) Misplaced labels are flagged as errors				      *
*									      *
*  Other changes are:							      *
*    a) IHex2Bin rewritten completely to accomodate all Intel .HEX features.  *
*    b) Hex2Bin rewritten to be a true free-form hex reader.		      *
*    d) 8051.H was altered so that CLR C is recognised, ACALL #07XXH is       *
*        handled correctly and ORL direct,#data printed properly	      *
*									      *
*  v2.0 user interface removed for compatibility with Unix etc, now runs      *
*    completely from command-line. Bug of non-printing interrupt comments     *
*    when addresses switched off fixed.					      *
*									      *
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "8051.h"


/******************************************************************************
* Global Variables                                                            *
******************************************************************************/

typedef unsigned char UBYTE;
typedef unsigned char BOOL;
#define TRUE    1
#define FALSE   0

#define MAX_JUMPS       1024
int no_jumps=0;
int *Jump;
BOOL *JumpDone;

int Location;           /* Of the NEXT byte to be read */

char TempFile[]="DIS8051.TMP";

unsigned int TopAddr;
BOOL Addresses;
BOOL NOPs;

char Title[]="Dis8051 v2.0, by Gary Benson <gbenson@iee.org>,";
char CopyR[]="1997-1998";


/******************************************************************************
* Pass 1 - Locate Jumps							      *
******************************************************************************/

int JumpAddr(int rel)
{
	int addr;

	addr=Location;
	if((rel<0)||(rel>255))
	{
		printf("Dis8051: Bad jump address\n");
		printf(" Location = ");
		if(Location<0x1000) putchar('0');
		if(Location<0x0100) putchar('0');
		if(Location<0x0010) putchar('0');
		printf("%x (decimal %d), rel=%d decimal\n",Location,Location,rel);
		return(EOF);
	}
	if((rel&0x80)==0x80)
	{
		addr-=128;
		rel&=0x7F;
	}
	if((rel<0)||(rel>127))
	{
		printf("Dis8051: Programmer error Number 4 (rel=%d) (ignore next message)\n  ",rel);
		return(EOF);
	}
	addr+=rel;
	if(addr<0)
	{
		printf("Dis8051: Programmer error Number 5 (addr=%d) (ignore next message)\n  ",addr);
		return(EOF);
	}
	return(addr);
}

int LabelNumber(int addr)
{
	int i,j;

	j=-1;
	for(i=0;i<no_jumps;i++)
	{
		if(addr==Jump[i]) j=i;
	}
	return(j);
}

int GetByte(FILE *Fh)
{
	int i;
	i=getc(Fh);
	Location++;
	return(i);
}

BOOL GetJumps(FILE *fH)
{
	int i,j,address;
	unsigned int swap;
	BOOL jump;
	UBYTE oc;

	Location=0;
	i=GetByte(fH);
	if(i==EOF)
	{
		printf("Dis8051: File contains no code, or is of incorrect format.");
		return(FALSE);
	}
	while((i!=EOF)&&(Location<=TopAddr))
	{
		/*
		* Check for relative jumps...
		*/
		jump=FALSE;
		oc=i;
		if(oc!=0)
		{
			if((oc&0x8F)==0) jump=TRUE;
			else if(oc==0x80) jump=TRUE;             /* SJMP */
			else if((oc>0xB3)&&(oc<0xC0)) jump=TRUE; /* CJNE */
			else if((oc>0xD7)&&(oc<0xE0)) jump=TRUE; /* DJNZ */
			else if(oc==0xD5) jump=TRUE;             /* DJNZ */
			else jump=FALSE;
		}

		/*
		* Calculate relative jump address
		*/
		if(jump)
		{
			/*
			* Skip intervening bytes
			*/
			if(oc<0x40) GetByte(fH);
			else if((oc>0xB3)&&(oc<0xD6)) GetByte(fH);
			i=GetByte(fH);
			address=JumpAddr(i);
			if(address==EOF)
			{
				printf("Dis8051: Unexpected end of file!");
				return(FALSE);
			}
		}

		/*
		* Calculate 11-bit Jump/Call address
		*/
		if((oc&0x0F)==0x01)
		{
			j=GetByte(fH);
			if(j==EOF)
			{
				printf("Dis8051: Unexpected end of file!");
				return(FALSE);
			}
			j&=0xFF;
			i=(int)(oc&0xE0)<<3;
			i+=j;
			address=i+(Location&0xF800);
			jump=TRUE;
		}

		/*
		* Calculate 16-bit Jump/Call address
		*/
		if((oc==0x02)||(oc==0x12))
		{
			i=GetByte(fH);
			j=GetByte(fH);
			if((i==EOF)||(j==EOF))
			{
				printf("Dis8051: Unexpected end of file!");
				return(FALSE);
			}
			i&=0xFF; j&=0xFF;
			i=i<<8;
			i+=j;
			address=i;
			jump=TRUE;
		}

		/*
		* If it is a jump, we have the address.
		* now, deal with this address...
		*/
		if(jump)
		{
			/*
			* Now see if a jump already goes there.
			*/
			j=LabelNumber(address);
			if(j==-1)
			{
				/*
				* This is a new jump point
				*/
				if(no_jumps==MAX_JUMPS)
				{
					printf("Dis8051: Too many jumps.");
					return(FALSE);
				}
				Jump[no_jumps]=address;
				no_jumps++;
			}
		}
		else
		{
			/*
			* Not a jump, so skip remaining bytes of operand
			*/
			i=oc%16;
			j=(oc-i)/16;
			if(arg1[j][i]!='-')
			{
				GetByte(fH);

				/*
				* Thats the 2nd byte got; is there another?
				*/
				if(arg1[j][i]=='L')
				{
					GetByte(fH);
				}
				else
				{
					j=-1;
					for(i=0;i<NUM_ARG2;i++) if(oc==arg2rows[i]) j=i;
					if(j!=-1) GetByte(fH);
				}
			}
		}

		i=GetByte(fH);
	}

	/*
	* Now bubble sort the jumps
	*/
	if(no_jumps>1)
	{
		for(i=0;i<(no_jumps-1);i++)
		{
			for(j=0;j<(no_jumps-1);j++)
			{
				if(Jump[j]>Jump[j+1])
				{
					swap=Jump[j];
					Jump[j]=Jump[j+1];
					Jump[j+1]=swap;
				}
			}
		}
	}
	return(TRUE);
}


/******************************************************************************
* Pass 2 - Disassemble							      *
******************************************************************************/

BOOL PrintArg(FILE *sFh,FILE *dFh,char Type,char oc)
{
	int i,j,k;

	switch(Type)
	{
		case 'D':                       /* 8-bit Data                   */
			i=GetByte(sFh);
			if(i==EOF) return(FALSE);
			if(i<0x10) fprintf(dFh,"0");
			fprintf(dFh,"%xH",i);
			break;
		case 'L':                       /* 16-bit data                  */
			i=GetByte(sFh);
			j=GetByte(sFh);
			if((i==EOF)||(j==EOF)) return(FALSE);
			i&=0xFF; j&=0xFF;
			i=i<<8;
			i+=j;
			if(i<0x1000) fprintf(dFh,"0");
			if(i<0x0100) fprintf(dFh,"0");
			if(i<0x0010) fprintf(dFh,"0");
			fprintf(dFh,"%xH",i);
			break;
		case 'B':                       /* Directly addressable bit     */
			i=GetByte(sFh);
			if(i==EOF) return(FALSE);
			if((i>127)&&(BitNames[i-128][0]!='-')) fprintf(dFh,"%s",BitNames[i-128]);
			else
			{
				if(i<0x10) fprintf(dFh,"0");
				fprintf(dFh,"%xH",i);
			}
			break;
		case 'A':
			i=GetByte(sFh);
			if(i==EOF) return(FALSE);
			k=i;
			j=-1;
			for(i=0;i<NUM_SFRS;i++)
			{
				if(SFRAddrs[i]==k) j=i;
			}
			if(j!=-1) fprintf(dFh,"%s",SFRNames[j]);
			else
			{
				if(k<0x10) fprintf(dFh,"0");
				fprintf(dFh,"%xH",k);
			}
			break;
		case 'R':
			i=GetByte(sFh);
			k=JumpAddr(i);
			if(k==EOF) return(FALSE);
			j=LabelNumber(k);
			if(j==-1)
			{
				printf("Dis8051: Programmer Error 2a (Loc=%d, jmpadd=%d)\n",Location,k);
				printf("* Label Dump:\n");
				for(i=0;i<no_jumps;i++) printf("\t%d:\t%d\n",i,Jump[i]);
				printf("\nIgnore the next error message...\n\n   ");
				return(FALSE);
			}
			fprintf(dFh,"LAB%d",j);
			break;
		case 'J':                       /* 11-bit address               */
			j=GetByte(sFh);
			if(j==EOF) return(FALSE);
			j&=0xFF;
			i=(int)(oc&0xE0)<<3;
			i+=j;
			k=i+(Location&0xF800);
			j=LabelNumber(k);
			if(j==-1)
			{
				printf("Dis8051: Programmer Error 2b (Loc=%d, jmpadd=%d)\n",Location,k);
				printf("* Label Dump:\n");
				for(i=0;i<no_jumps;i++) printf("\t%d:\t%d\n",i,Jump[i]);
				printf("\nIgnore the next error message...\n\n   ");
				return(FALSE);
			}
			fprintf(dFh,"LAB%d",j);
			break;
		case 'K':                       /* 16-bit address               */
			i=GetByte(sFh);
			j=GetByte(sFh);
			if((i==EOF)||(j==EOF)) return(FALSE);
			i&=0xFF; j&=0xFF;
			i=i<<8;
			i+=j;
			j=LabelNumber(i);
			if(j==-1)
			{
				printf("Dis8051: Programmer Error 2c (Loc=%d, jmpadd=%d)\n",Location,i);
				printf("* Label Dump:\n");
				for(i=0;i<no_jumps;i++) printf("\t%d:\t%d\n",i,Jump[i]);
				printf("\nIgnore the next error message...\n\n   ");
				return(FALSE);
			}
			fprintf(dFh,"LAB%d",j);
			break;
		default:
			printf("Dis8051: Programmer Error 3 (Ignore the next error message)\n  ");
			return(FALSE);
	}
	return(TRUE);
}

BOOL DisAss(FILE *sFh,FILE *dFh)
{
	int Warnings=0;
	int JumpWarns=0;
	int Lines=0;
	int len;
	UBYTE lastoc=1,oc=1;
	int i,j,Address;

	rewind(sFh);
	Location=0;
	fprintf(dFh,";\n;\t Disassembled by:\n;\t %s\n",Title);
	fprintf(dFh,";\t  (C) Copyright %s, All rights reserved.\n;\n\n",CopyR);

	i=GetByte(sFh);
	while((i!=EOF)&&(Location<=TopAddr))
	{
		lastoc=oc;
		oc=i;

		/*
		* Is this location a jump point?
		*/
		j=-1;
		for(i=0;i<no_jumps;i++) if((Location-1)==Jump[i]) j=i;

		if((!NOPs)&&(oc==0x00)&&(j==-1))
		{
			if(lastoc!=0x00) putc('\n',dFh);
		}
		else
		{
			/*
			* Print the address
			*/
			Address=Location-1;
			if(Addresses)
			{
				fprintf(dFh,"%04X:\t",Address);
			}

			/*
			* Is this location a jump point?
			* If so, print the label.
			* and set lastoc in case we are on a NOP.
			*/
			if(j!=-1)
			{
				fprintf(dFh,"LAB%d:",j);
				JumpDone[j]=TRUE;
				lastoc=0xFF;
			}

			/*
			* Print the mnemonic, and the first argument if necessary
			*/
			len=strlen(Instr[oc]);
			fprintf(dFh,"\t%s",Instr[oc]);
			i=oc%16;
			j=(oc-i)/16;

			switch(arg1[j][i])
			{
				case 'W':
					fprintf(dFh,"\t; WARNING: This is an undefined operation");
					Warnings++;
					break;
				case '-':
					break;
				case 'L':
				case 'K':
				case 'R':
				case 'J':
					len+=2;
				case 'B':
				case 'A':
				case 'D':
					len+=3;
					if(!PrintArg(sFh,dFh,arg1[j][i],oc))
					{
						printf("Dis8051: Unexpected end of file.");
						return(FALSE);
					}
					break;
				default:
					printf("Dis8051: Programmer Error 1.");
					return(FALSE);
			}

			/*
			* Has it got extra text?
			*/
			j=-1;
			for(i=0;i<NUM_EXTRAS;i++) if(oc==exrows[i]) j=i;
			if(j!=-1) fprintf(dFh,"%s",extra[j]);

			/*
			* Has it got another argument?
			*/
			j=-1;
			for(i=0;i<NUM_ARG2;i++) if(oc==arg2rows[i]) j=i;

			if(j!=-1)
			{
				len+=3;
				if(!PrintArg(sFh,dFh,arg2[j],oc))
				{
					printf("Dis8051: Unexpected end of file.");
					return(FALSE);
				}
			}
			Lines++;

			/*
			* Print the vectors
			*/
			if(Address==0x0000)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; Reset Vector");
			}
			else if(Address==0x0003)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; External Interrupt 0");
			}
			else if(Address==0x000B)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; Timer Interrupt 0");
			}
			else if(Address==0x0013)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; External Interrupt 1");
			}
			else if(Address==0x001B)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; Timer Interrupt 1");
			}
			else if(Address==0x0023)
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; Serial Interrupt");
			}
			else if((Address==0x002B)||(Address==0x0033)||(Address==0x003B))
			{
				if(len<8) putc('\t',dFh);
				fprintf(dFh,"\t; Unused Interrupt");
			}

			/*
			* Finish off the line.
			*/
			fprintf(dFh,"\n");
			if(ferror(dFh))
			{
				printf("Dis8051: Write error in destination file.");
				return(FALSE);
			}
		}

		i=GetByte(sFh);
	}

	/*
	* Print a report, ie: X bytes of machine code were
	* disassembled into Y lines of source code; There
	* were Z warnings encountered.
	*/
	printf("Disassembly Successful\n\n");
	printf(" %d bytes of hex were disassembled to %d lines of source code.\n ",Location-1,Lines);

	if(Warnings==0) printf("There were no undefined operations.");
	else if(Warnings==1) printf("There was one undefined operation.");
	else printf("There were %d undefined operations.",Warnings);
	printf("\n ");

	/*
	* Print missed labels warning
	*/
	for(i=0;i<no_jumps;i++)
	{
		if(!JumpDone[i])
		{
			JumpWarns++;
			j=i;
		}
	}
	if(JumpWarns==0) printf("There were no missing labels.");
	else if(JumpWarns==1)
	{
		printf("There was one missing jump, LAB%d, at address %04XH\n",j,Jump[j]);
	}
	else
	{
		printf("The following labels are missing:\n\n");
		for(i=0;i<no_jumps;i++)
		{
			if(!JumpDone[i])
			{
				printf("\tLAB%d\t(%04XH)\n",i,Jump[i]);
			}
		}
	}
	return(TRUE);
}


/******************************************************************************
* Other Functions                                                             *
******************************************************************************/

int GetULimit(char *Line)
{
	int Byte;

	if(Line==NULL)	   return(0xFFFF);
	if(strlen(Line)<1) return(0xFFFF);

	sscanf(Line,"%x",&Byte);
	return(Byte);
}

char *OutputFilename(char *InputFile)
{
	char *OutputFile;
	int i,j;

	j=strlen(InputFile);
	if(j==0)
	{
		printf("Dis8051: Invalid filename\n");
		return(NULL);
	}

	for(i=j-1;i>=0;i--)
	{
		if(InputFile[i]=='.')
		{
			j=i;
			i=-1;
		}
	}
	OutputFile=calloc(j+5,sizeof(char));
	if(OutputFile==NULL)
	{
		printf("Dis8051: Can't allocate a few bytes of RAM!!!\n");
		return(NULL);
	}

	strncpy(OutputFile,InputFile,j);
	OutputFile[j+0]='.';
	OutputFile[j+1]='a';
	OutputFile[j+2]='s' ;
	OutputFile[j+3]='m';
	OutputFile[j+4]= 0 ;

	return(OutputFile);
}

char Command[BUFSIZ];

main(int argc,char *argv[])
{
	FILE *srce_Fh,*dest_Fh;
	char *sFName,*dFName;
	BOOL Success=FALSE;
	int i;

	/*
	* Check the command line.
	*/
	printf("\n%s\n",Title);
	if(argc!=3)
	{
		printf(" (C) Copyright %s Gary Benson, All Rights Reserved.\n",CopyR);
		printf("\nFormat: Dis8051 <options> <source_file>");
		printf("\n where options is of the form <Y|N><Y|N>[len]");
		printf("\n  <Y|N> Print addresses (yes or no)");
		printf("\n  <Y|N> Print NOP's (yes or no)");
		printf("\n  [len] Length to disassemble (in hex), Default=FFFF\n");
		printf("\n A valid command would be Dis8051 YN4000 filename.ext");
		printf("\n If the file has the extension .HEX, it will be treated");
		printf("\n  as either Intel Hex or freeform hex depending on what");
		printf("\n  the format of the file actually is.\n\n");

		exit(EXIT_FAILURE);
	}

	/*
	* Parse the command line.
	*/
	if(toupper(argv[1][0]=='Y'))
	{
		Addresses=TRUE;
	}
	else if(toupper(argv[1][0]=='N'))
	{
		Addresses=FALSE;
	}
	else
	{
		printf("Dis8051: Invalid options string %s.\n",argv[1]);
		exit(EXIT_FAILURE);
	}

	if(toupper(argv[1][1]=='Y'))
	{
		NOPs=TRUE;
	}
	else if(toupper(argv[1][1]=='N'))
	{
		NOPs=FALSE;
	}
	else
	{
		printf("Dis8051: Invalid options string %s.\n",argv[1]);
		exit(EXIT_FAILURE);
	}
	TopAddr=GetULimit(&argv[1][2]);
	printf("\nTop Address set to %04XH\n",TopAddr);

	sFName=argv[2];
	dFName=OutputFilename(sFName);
	if(dFName==NULL)
	{
		exit(EXIT_FAILURE);
	}


	/*
	* Preprocess: Convert the file into a binary image if necessary.
	*/
#ifdef UNIX
	if(!strcasecmp(&sFName[strlen(sFName)-4],".hex"))
#else
	if(!stricmp(&sFName[strlen(sFName)-4],".hex"))
#endif
	{
		srce_Fh=fopen(sFName,"r");
		if(srce_Fh==NULL)
		{
			printf("Dis8051: Can't open file %s.",sFName);
			exit(EXIT_FAILURE);
		}
		i=getc(srce_Fh);
		fclose(srce_Fh);

		if(i==EOF)
		{
			printf("Dis8051: File %s contains no data.",sFName);
			exit(EXIT_FAILURE);
		}

		sprintf(Command,"ihex2bin %s %s",sFName,TempFile);
		printf("Preprocessor: Converting to binary image...\n ");
		if(i==':') system(Command);
		else system(&Command[1]);

		sFName=TempFile;

		srce_Fh=fopen(TempFile,"rb");
		if(srce_Fh) fclose(srce_Fh);
		else
		{
			printf("Dis8051: Can't convert file.");
			exit(EXIT_FAILURE);
		}
	}


	/*
	* The main disassembly.
	*/
	Jump=(int *)calloc(MAX_JUMPS,2);
	if(Jump)
	{
		JumpDone=(BOOL *)calloc(MAX_JUMPS,1);
		if(JumpDone)
		{
			srce_Fh=fopen(sFName,"rb");
			if(srce_Fh)
			{
				printf("Pass 1: Locating Jumps...\n");
				if(GetJumps(srce_Fh))
				{
					printf("Pass 2: Disassembling...\n");
					dest_Fh=fopen(dFName,"w");
					if(dest_Fh)
					{
						Success=DisAss(srce_Fh,dest_Fh);
						fclose(dest_Fh);
					}
					else printf("Dis8051: Can't write file %s.",dFName);
				}
				fclose(srce_Fh);
			}
			else printf("Dis8051: Can't open file %s.",sFName);
			free(JumpDone);
		}
		else printf("Dis8051: Can't allocate memory.");
		free(Jump);
	}
	else printf("Dis8051: Can't allocate memory.");

	remove(TempFile);
	printf("\n\n");
	exit(EXIT_SUCCESS);
}


/******************************************************************************
* End of DIS8051.C							      *
******************************************************************************/
