/******************************************************************************
* IHex2Bin v2.0								      *

  *****************************************
  * Specification of the Intel-HEX Format *
  *****************************************

This object file format is supported by many cross assemblers, utilities,
and most EPROM blowers.
An Intel-HEX file is a 7-bit ASCII text file, that contains a sequence of
data records and an end record. Every record is a line of text that starts
with a colon and ends with CR and LF.
Data records contain up to 16 data bytes, a 16-bit load address, a record
type byte and an 8-bit checksum. All numbers are represented by upper case
ASCII-hex characters.

       DATA RECORD:

       Byte 1               colon (:)
            2 and 3         number of binary data bytes for this record
            4 and 5         load address for this record, high byte
            6 and 7         load address  "    "     "     low byte
            8 and 9         record type: 00  (data record)
           10 to  x         data bytes, two characters each
          x+1 to  x+2       checksum (two characters)
          x+3 to  x+4       CR and LF

       A typical data record looks like

                         :10E0000002E003E4F588758910F58DF58BD28E302A

The end record is the last line of the file.
In principle it is structured like a data record, but the number of data
bytes is 00, the record type is 01 and the load-address field is 0000.

       END RECORD:

       Byte 1               colon (:)
            2 and 3         00   (number of data bytes)
            4 and 5         00   (load address, high byte)
            6 and 7         00   (load address, low byte)
            8 and 9         record type: 01  (end record)
           10 and 11        checksum (two characters)
           12 and 13        CR and LF


       The typical END record looks like

                         :00000001FF

The checksum is the two's complement of the 8-bit sum, without carry, of
the byte count, the two load address bytes, the record type byte and all
data bytes.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/******************************************************************************
* Global Variables							      *
******************************************************************************/

typedef unsigned char UBYTE;
typedef unsigned char BOOL;
#define TRUE	1
#define FALSE	0

#define DATA_SIZE	0x20000		/* A maximum size of 128k (arbitary) */
#define DATA_RECORD	0x00
#define END_RECORD	0x01


/******************************************************************************
* Functions 								      *
******************************************************************************/

int Hex(char *Line)
{
	unsigned char value[]="0123456789ABCDEF";
	int i,Byte;
	char nybble=0xF0;

	for(i=0;i<16;i++) if(toupper(Line[0])==value[i]) nybble=i;
	if(nybble==0xF0) return(EOF);
	Byte=nybble*16;

	nybble=0xF0;
	for(i=0;i<16;i++) if(toupper(Line[1])==value[i]) nybble=i;
	if(nybble==0xF0) return(EOF);
	Byte+=nybble;

	return(Byte);
}

int ReadIHex(FILE *srce_Fh,char *Data)
{
	UBYTE Buffer[BUFSIZ];
	UBYTE buf[BUFSIZ];
	int i,j,linelen;
	int Addr,length=0;
	char *c;

	UBYTE reclen;
	UBYTE AddrL,AddrH;
	UBYTE rectype;
	char *CSum;
	char S;

	c=fgets(Buffer,BUFSIZ,srce_Fh);
	if(c==NULL)
	{
		printf("IHex2Bin: File contains no data.");
		return(0);
	}
	while(c!=NULL)
	{
		/*
		* Chop of CR/LF
		*/
		linelen=strlen(Buffer);
		if(linelen>2)
		{
			if(Buffer[linelen-1]==0x0A) Buffer[linelen-1]=0;
			if(Buffer[linelen-1]==0x0D) Buffer[linelen-1]=0;
			if(Buffer[linelen-2]==0x0A) Buffer[linelen-2]=0;
			if(Buffer[linelen-2]==0x0D) Buffer[linelen-2]=0;
		}
		linelen=strlen(Buffer);

		/*
		* Check the line starts with a colon
		*/
		if(Buffer[0]!=':')
		{
			printf("IHex2Bin: Invalid line :\n%s",Buffer);
			return(0);
		}

		/*
		* Decode the identifiers and the address
		*/
		for(i=0;i<(linelen-1)/2;i++)
		{
			j=Hex(&Buffer[1+i*2]);
			if(j<0)
			{
				printf("IHex2Bin: Invalid line at byte %d:\n%s",i,Buffer);
				return(0);
			}
			else buf[i]=j;
		}
		reclen=buf[0];
		rectype=buf[3];

		AddrL=buf[2]; AddrH=buf[1];
		Addr=(int)AddrL+(((int)AddrH)<<8);
		if((Addr+reclen)>=DATA_SIZE)
		{
			printf("IHex2Bin: This version only supports %d bytes of data\n");
			return(0);
		}

		if(rectype==DATA_RECORD)
		{
			/*
			* Check the checksum
			*/
			CSum=&buf[4+reclen];
                        S=AddrH+AddrL+reclen+rectype;
			for(i=0;i<reclen;i++) S+=buf[4+i];
			S=-S;
			if(S!=*CSum)
			{
				printf("IHex2Bin: Checksum error on line:\n%s",Buffer);
				printf("  Read in %d\tCalculated %d",*CSum,S);
				return(0);
			}

			/*
			* Checksum is ok, so copy data into array.
			*/
			for(i=0;i<reclen;i++)
			{
				Data[Addr+i]=buf[4+i];
			}

			if((Addr+reclen)>length) length=Addr+reclen;
			c=fgets(Buffer,BUFSIZ,srce_Fh);
		}
		else if(rectype==END_RECORD) return(length);
		else
		{
			printf("IHex2Bin: Unknown identifier.");
			return(0);
		}
	}
	printf("IHex2Bin: Unexpected end of file (No end record found).");
	return(0);
}

main(int argc,char *argv[])
{
	FILE *srce_Fh,*dest_Fh;
	BOOL Success=FALSE;
	char *Data;
	int Length;

	if(argc!=3)
	{
		printf("\nIHex2Bin v2.0, by Gary Benson.");
		printf("\n (C) Copyright 1997-1998 Gary Benson, All Rights Reserved.\n");
		printf("\nFormat: IHex2Bin <source_file> <dest_file>\n\n");

		exit(EXIT_FAILURE);
	}

	Data=calloc(DATA_SIZE,1);
	if(Data)
	{
		srce_Fh=fopen(argv[1],"r");
		if(srce_Fh)
		{
			dest_Fh=fopen(argv[2],"wb");
			if(dest_Fh)
			{
				Length=ReadIHex(srce_Fh,Data);
				if(Length>0)
				{
					fwrite(Data,Length,1,dest_Fh);
					if(ferror(dest_Fh))
					{
						printf("IHex2Bin: Error writing file %s.",argv[2]);
					}
					else Success=TRUE;
				}
				fclose(dest_Fh);
			}
			else printf("IHex2Bin: Can't open write file %s.",argv[2]);

			fclose(srce_Fh);
		}
		else printf("IHex2Bin: Can't open file %s.",argv[1]);
		free(Data);
	}
	else printf("IHex2Bin: Can't allocate memory.");

	if(!Success)
	{
		remove(argv[2]);
		exit(EXIT_FAILURE);
	}
	else exit(EXIT_SUCCESS);
}


/******************************************************************************
* End of IHEX2BIN.C							      *
******************************************************************************/
